"""PyECDAR, a tool for timed games and timed specifications.
Copyright (C) INRIA Rennes - Louis-Marie Traonouez

Louis-Marie.Traonouez@inria.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""
    
import pyuppaal
import re
import copy
import logging

logging.basicConfig()
logger = logging.getLogger('MODEL_TRANSFORM')


def contraryConstraint(cons):
  """Return the contrary of a simple rectangular guard:
  e.g. x <= k --> x > k, where k is a positive integer."""
  if not cons:
    return ""
  assert(isinstance(cons,str))
  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff>\d+)\s*$',cons)
  if r:
    if r.group("sign") == "<":
      return r.group("clock") + '>=' + r.group('coeff')
    elif r.group("sign") == "<=":
      return r.group("clock") + '>' + r.group('coeff')
    elif r.group("sign") == ">":
      return r.group("clock") + '<=' + r.group('coeff')
    elif r.group("sign") == ">=":
      return r.group("clock") + '<' + r.group('coeff')
    else:
      return r.group("clock") + '<' + r.group('coeff') + ' && ' + r.group("clock") + '>' + r.group('coeff')
      
  r = re.search('^\s*(?P<coeff>\d+)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if r.group("sign") == ">":
      return r.group("clock") + '>=' + r.group('coeff')
    elif r.group("sign") == ">=":
      return r.group("clock") + '>' + r.group('coeff')
    elif r.group("sign") == "<":
      return r.group("clock") + '<=' + r.group('coeff')
    elif r.group("sign") == "<=":
      return r.group("clock") + '<' + r.group('coeff')
    else:
      return r.group("clock") + '<' + r.group('coeff') + ' && ' + r.group("clock") + '>' + r.group('coeff')
  logger.error("can't get contrary constraint: "+cons)
  return ""
  

def enlargedConstraint(cons,delta):
  """Enlarge a simple rectangular constraint by delta:
  e.g. x <= k --> x <= k+delta (where delta is either an integer or a parameter).
  
  Return the enlarge constraint."""
  if not cons:
    return ""
  assert(isinstance(cons,str))
  assert(isinstance(delta,str) | isinstance(delta,int))
  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff>\d+)\s*$',cons)
  if r:
    if r.group("sign") == "<":
      if isinstance(delta,int):
	return r.group("clock") + '<' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == ">":
      if isinstance(delta,int):
	return r.group("clock") + '>' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
    else: # sign is == : The constraint is split into 2 inequalities.
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta) + ' && ' + r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)	  
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta) + ' && ' + r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
	
  r = re.search('^\s*(?P<coeff>\d+)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if r.group("sign") == ">":
      if isinstance(delta,int):
	return r.group("clock") + '<' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)
    elif r.group("sign") == "<":
      if isinstance(delta,int):
	return r.group("clock") + '>' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)
      else:
	return r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
    else: # sign is == : The constraint is split into 2 inequalities.
      if isinstance(delta,int):
	return r.group("clock") + '<=' + str(int(r.group('coeff'))+delta) + ' && ' + r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)	  
      else:
	return r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta) + ' && ' + r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)
  logger.error("can't enlarge constraint: "+cons)
  return ""
  
  
def restrictedConstraint(cons,delta,upper=False):
  """Reduced a simple rectangular constraint by delta:
  e.g. x <= k --> x <= k-delta (where delta is either an integer or a parameter).
  Return a list with the reduced constraint and its contrary. 
  If equality return an empty list."""
  if not cons:
    return []
  assert(isinstance(cons,str))
  assert(isinstance(delta,str) | isinstance(delta,int))
  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff>\d+)\s*$',cons)
  if r:
    if upper and  (r.group("sign") == ">" or r.group("sign") == ">="):
      return [cons,cons]
      
    if r.group("sign") == "<":
      if isinstance(delta,int):
	return [r.group("clock") + '<' + str(int(r.group('coeff'))-delta), r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return [r.group("clock") + '<=' + str(int(r.group('coeff'))-delta), r.group("clock") + '>' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<=' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == ">":
      if isinstance(delta,int):
	return [r.group("clock") + '>' + str(int(r.group('coeff'))+delta), r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)]
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return [r.group("clock") + '>=' + str(int(r.group('coeff'))+delta), r.group("clock") + '<' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>=' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)]
    else:
      return []
  r = re.search('^\s*(?P<coeff>\d+)\s*(?P<sign>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if upper and  (r.group("sign") == "<" or r.group("sign") == "<="):
      return [cons,cons]
    
    if r.group("sign") == ">":
      if isinstance(delta,int):
	return [r.group("clock") + '<' + str(int(r.group('coeff'))-delta), r.group("clock") + '>=' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>=' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == ">=":
      if isinstance(delta,int):
	return [r.group("clock") + '<=' + str(int(r.group('coeff'))-delta), r.group("clock") + '>' + str(int(r.group('coeff'))-delta)]
      else:
	return [r.group("clock") + '<=' + r.group('coeff') +'-'+ str(delta), r.group("clock") + '>' + r.group('coeff') +'-'+ str(delta)]
    elif r.group("sign") == "<":
      if isinstance(delta,int):
	return [r.group("clock") + '>' + str(int(r.group('coeff'))+delta), r.group("clock") + '<=' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<=' + r.group('coeff') +'+'+ str(delta)]
    elif r.group("sign") == "<=":
      if isinstance(delta,int):
	return [r.group("clock") + '>=' + str(int(r.group('coeff'))+delta), r.group("clock") + '<' + str(int(r.group('coeff'))+delta)]
      else:
	return [r.group("clock") + '>=' + r.group('coeff') +'+'+ str(delta), r.group("clock") + '<' + r.group('coeff') +'+'+ str(delta)]
    else:
      return []
  logger.error("can't restrict constraint: "+cons)
  return []

def scaledConstraint(cons,factor):
  """Scaled the constant coefficient in a simple guard by a factor:
  e.g. x <= k --> x <= k*factor
        x <= k + l*d --> x <= k*factor + l*d """
  if not cons:
    return ""
  assert(isinstance(cons,str))
  assert(isinstance(factor,int))

  r = re.search('^\s*(?P<clock>[a-zA-Z_]\w*)\s*(?P<sign1>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<coeff1>(-)?\s*\d+)?\s*(?P<sign2>((-)|(\+)))?\s*(?P<coeff2>\d+)?\s*(?P<sign3>(\*))?\s*(?P<delta>\w+)?\s*$',cons)
  coeff_str=""
  if r:
    if r.group("coeff1"):
      if not r.group("delta") or r.group("sign2"):
	coeff = int(r.group("coeff1"))*factor
	coeff_str +=  str(coeff)
      else:
	coeff = int(r.group("coeff1"))
	coeff_str += str(coeff)
    if r.group("sign2"):
      coeff_str += r.group("sign2")
    if r.group("coeff2"):
      coeff = int(r.group("coeff2"))
      coeff_str += str(coeff)
    if r.group("sign3"):
      coeff_str += r.group("sign3")
    if r.group("delta"):
      coeff_str += r.group("delta")
    return r.group("clock") + r.group("sign1") + coeff_str
    
  r = re.search('^\s*(?P<coeff1>(-)?\s*\d+)?\s*(?P<sign2>((-)|(\+)))?\s*(?P<coeff2>\d+)?\s*(?P<sign3>(\*))?\s*(?P<delta>\w+)?\s*(?P<sign1>((>=)|(>)|(==)|(<=)|(<)))\s*(?P<clock>[a-zA-Z_]\w*)\s*$',cons)
  if r:
    if r.group("coeff1"):
      if not r.group("delta") or r.group("sign2"):
	coeff = int(r.group("coeff1"))*factor
	coeff_str += str(coeff)
      else:
	coeff = int(r.group("coeff1"))
	coeff_str += str(coeff)
    if r.group("sign2"):
      coeff_str += r.group("sign2")
    if r.group("coeff2"):
      coeff = int(r.group("coeff2"))
      coeff_str += str(coeff)
    if r.group("sign3"):
      coeff_str += r.group("sign3")
    if r.group("delta"):
      coeff_str += r.group("delta")
    return coeff_str + r.group("sign1") + r.group("clock")
  logger.error("can't scale constraint: "+cons)
  return cons
 

def scaledTemplateConstants(template,factor):
  """Scale the constants coefficient in a rectangular automata.  
  Return a new modified template."""
  assert(isinstance(factor,int))
  assert(isinstance(template,pyuppaal.Template))
  
  scaled_template = copy.deepcopy(template)
  scaled_template.name += "S"+str(factor)
  
  revert_dbmpyuppaal(scaled_template)

  for l in scaled_template.locations:
    invl = str(l.invariant).split("&&")
    for i in range(len(invl)):
      invl[i] = scaledConstraint(invl[i],factor)
    l.invariant.value = ' && '.join(invl)
      
  for t in scaled_template.transitions:
    guardl = str(t.guard).split("&&")
    for i in range(len(guardl)):
      guardl[i] = scaledConstraint(guardl[i],factor)
    t.guard.value = ' && '.join(guardl)
  
  return scaled_template
    

def implemPerturbation(template,delta):
  """Add a delay perturbation delta to the guards of the template,
  (where delta is either an integer or a parameter).
  Return a new template for the perturbed implementation."""
  rob_impl = copy.deepcopy(template)
  rob_impl.name += "Rob"+str(delta)
  
  revert_dbmpyuppaal(rob_impl)
  
  lu = None
  new_lu = False
  output_actions = []
  input_actions = []
  lu_transitions = []
  
  for l in rob_impl.locations:
    if l.name.value == "UNIV":
      lu = l
    invl = str(l.invariant).split("&&")
    for i in range(len(invl)):
      invl[i] = enlargedConstraint(invl[i],delta)
    l.invariant.value = ' && '.join(invl)
  if not lu:
    new_lu = True
    lu = pyuppaal.Location(name="UNIV")
    rob_impl.locations.append(lu)    
      
  for t in rob_impl.transitions:
    if t.controllable is False:  	# output transition
      if new_lu and str(t.synchronisation) and str(t.synchronisation) not in output_actions:
	output_actions.append(str(t.synchronisation))
      guardl = str(t.guard).split("&&")
      for i in range(len(guardl)):
	guardl[i] = enlargedConstraint(guardl[i],delta)
      t.guard.value = ' && '.join(guardl)
     
    else:	# input transition
      if new_lu and str(t.synchronisation) and str(t.synchronisation) not in input_actions:
	input_actions.append(str(t.synchronisation))
      guardl = str(t.guard).split("&&")
      guardl_rob = []
      for i in range(len(guardl)):
	cons_robl = restrictedConstraint(guardl[i],delta)
	if cons_robl:
	  guardl_rob.append(cons_robl[0])
	  guard_lu=guardl[:]
	  guard_lu.append(cons_robl[1])
	  lu_transitions.append(pyuppaal.Transition(source=t.source, \
			target = lu, \
			synchronisation = str(t.synchronisation), \
			guard = ' && '.join(guard_lu)))
      t.guard.value = ' && '.join(guardl_rob)
  
  if new_lu:
    for action in output_actions:
      rob_impl.transitions.append(pyuppaal.Transition(source=lu, \
	      target = lu, \
	      controllable = False, \
	      synchronisation = action))
    for action in input_actions:
      rob_impl.transitions.append(pyuppaal.Transition(source=lu, \
	      target = lu, \
	      controllable = True, \
	      synchronisation = action))	      
     
  rob_impl.transitions.extend(lu_transitions)
  return rob_impl
  

  
def robustTGA(template,delta,bad_loc_name="BAD"):
  """Compute the robust TGA in which one player wins iff it has a delta-robust strategy.
  Delta can be either a value or parameter.
  Controllable transitions must be assigned to the desired player.
  Return a new template for the robust TGA."""
  rob_game = copy.deepcopy(template)
  rob_game.name += "RobTG"
  rob_game.declaration += "\nclock robClk;"
  
  revert_dbmpyuppaal(rob_game)
  
  bad_loc = None
  for l in rob_game.locations:
    if str(l.name) == bad_loc_name:
      bad_loc = l
      break;
  if not bad_loc:
    bad_loc = pyuppaal.Location(name=bad_loc_name)
    rob_game.locations.append(bad_loc)
  
  new_transitions = []
  removed_transitions = []
  
  for t in rob_game.transitions:
    if t.controllable:
      if re.search('(==)',str(t.guard)):
	removed_transitions.append(t) # if the guard is an equality the transition is permanently removed since it cannot be robustly taken
        continue
	
      source_loc = t.source
      target_loc = t.target
      alpha_loc = pyuppaal.Location(name=str(source_loc.name)+"_"+str(t.synchronisation)[:-1]+"_"+str(target_loc.name)+"_A")
      beta_loc = pyuppaal.Location(name=str(source_loc.name)+"_"+str(t.synchronisation)[:-1]+"_"+str(target_loc.name)+"_B")
      alpha_loc.invariant.value = "robClk <= "+str(delta)
      rob_game.locations.append(alpha_loc)
      rob_game.locations.append(beta_loc)
      
      guardl = str(t.guard).split("&&")
      guardl.extend(str(source_loc.invariant).split("&&"))
      for guardi in guardl:
	if re.search("<=|<",guardi):
	  new_transitions.append(pyuppaal.Transition(source=alpha_loc, \
		target = bad_loc, \
		controllable = False, \
		guard = contraryConstraint(guardi)+" && robClk <= "+str(delta), \
		assignment="robClk=0"))
	  new_transitions.append(pyuppaal.Transition(source=beta_loc, \
		target = bad_loc, \
		controllable = False, \
		guard = contraryConstraint(guardi)+" && robClk <= "+str(delta), \
		assignment="robClk=0"))
           
      tassignment = str(t.assignment)
      if tassignment:
	tassignment += ",robClk=0"
      else:
	tassignment += "robClk=0"
      new_transitions.append(pyuppaal.Transition(source=alpha_loc, \
	      target = beta_loc, \
	      controllable = True, \
	      guard = "robClk == "+str(delta), \
	      assignment="robClk=0"))
      
      new_transitions.append(pyuppaal.Transition(source=alpha_loc, \
	      target = target_loc, \
	      controllable = False, \
	      guard = "robClk <= "+str(delta), \
	      assignment=tassignment))
	      
      new_transitions.append(pyuppaal.Transition(source=beta_loc, \
	      target = target_loc, \
	      controllable = False, \
	      guard = "robClk <= "+str(delta), \
	      assignment=tassignment))
      
      t.target = alpha_loc
      t.assignment.value="robClk=0"
      
      for u in rob_game.transitions:
	if not u.controllable and u.source==source_loc and u.target != u.source: #self-loop with input sync are not added to the robust moved of the spoiler
	  uguard = str(u.guard)
	  if uguard:
	    uguard += " && robClk <= "+str(delta)
	  else:
	    uguard = "robClk <= "+str(delta)
	  uassignment = str(u.assignment)
	  if uassignment:
	    if not re.search("robClk=0",uassignment):
	      uassignment += ",robClk=0"
	  else:
	    uassignment = "robClk=0"
	  found = False
	  if u.target == bad_loc:
	    for t in new_transitions: #we search for equivalent transitions, which happen specifically if we apply the transformations: consistencyGame and robustTGA successively.  
	      if (t.source == alpha_loc or t.source == beta_loc) and \
		  t.target == bad_loc and \
		  str(t.guard) == uguard and \
		  str(t.assignment) == uassignment:
		found = True
		break
	  if not found:  
	    new_transitions.append(pyuppaal.Transition(source=alpha_loc, \
		    target = u.target, \
		    controllable = False, \
		    guard = uguard, \
		    synchronisation = str(u.synchronisation), \
		    assignment=uassignment))
	    
	    new_transitions.append(pyuppaal.Transition(source=beta_loc, \
		    target = u.target, \
		    controllable = False, \
		    guard = uguard, \
		    synchronisation = str(u.synchronisation), \
		    assignment=uassignment))
    else:
      if t.target != t.source:
	if str(t.assignment):
	  t.assignment.value += ", robClk=0"
	else:
	  t.assignment.value = "robClk=0"
      
  rob_game.transitions.extend(new_transitions)
  for t in removed_transitions:
    rob_game.transitions.remove(t)
     
  return rob_game  
 

# Implementation of robust game transformation from FORMATS
# does not work with parametric analysis
# 
def robustTGAFormats(template,delta,bad_loc_name="BAD"):
  """Compute the robust TGA in which one player wins iff it has a delta-robust strategy.
  Delta can be either a value or parameter.
  Controllable transitions must be assigned to the desired player.
  Return a new template for the robust TGA."""
  rob_game = copy.deepcopy(template)
  rob_game.name += "RobTG"
  rob_game.declaration += "\nclock robClk;"
  
  if isinstance(delta,str):
    ddelta = "2*"+delta
  else:
    ddelta = 2*delta
  
  revert_dbmpyuppaal(rob_game)
  
  bad_loc = None
  for l in rob_game.locations:
    if str(l.name) == bad_loc_name:
      bad_loc = l
      break;
  if not bad_loc:
    bad_loc = pyuppaal.Location(name=bad_loc_name)
    rob_game.locations.append(bad_loc)
  
  new_transitions = []
  removed_transitions = []
  
  for t in rob_game.transitions:
    if t.controllable:
      if re.search('(==)',str(t.guard)):
	removed_transitions.append(t) # if the guard is an equality the transition is permanently removed since it cannot be robustly taken
        continue
	
      source_loc = t.source
      target_loc = t.target
      alpha_loc = pyuppaal.Location(name=str(source_loc.name)+"_"+str(t.synchronisation)[:-1]+"_"+str(target_loc.name)+"_A")
      beta_loc = pyuppaal.Location(name=str(source_loc.name)+"_"+str(t.synchronisation)[:-1]+"_"+str(target_loc.name)+"_B")
      alpha_loc.invariant.value = "robClk <= "+str(delta)
      rob_game.locations.append(alpha_loc)
      rob_game.locations.append(beta_loc)
      
      guardl = str(t.guard).split("&&")
      guardl_rob = []
      for i in range(len(guardl)):
	cons_robl = restrictedConstraint(guardl[i],ddelta,True)
	if cons_robl:
	  guardl_rob.append(cons_robl[0])
      t.guard.value = ' && '.join(guardl_rob)
      
      tassignment = str(t.assignment)
      if tassignment:
	tassignment += ",robClk=0"
      else:
	tassignment += "robClk=0"
      new_transitions.append(pyuppaal.Transition(source=alpha_loc, \
	      target = beta_loc, \
	      controllable = True, \
	      guard = "robClk == "+str(delta), \
	      assignment="robClk=0"))
      
      new_transitions.append(pyuppaal.Transition(source=alpha_loc, \
	      target = target_loc, \
	      controllable = False, \
	      guard = "robClk <= "+str(delta), \
	      assignment=tassignment))
	      
      new_transitions.append(pyuppaal.Transition(source=beta_loc, \
	      target = target_loc, \
	      controllable = False, \
	      guard = "robClk <= "+str(delta), \
	      assignment=tassignment))
      
      t.target = alpha_loc
      t.assignment.value="robClk=0"
      
      for u in rob_game.transitions:
	if not u.controllable and u.source==source_loc and u.target != u.source: #self-loop with input sync are not added to the robust moved of the spoiler
	  uguard = str(u.guard)
	  if uguard:
	    uguard += " && robClk <= "+str(delta)
	  else:
	    uguard = "robClk <= "+str(delta)
	  uassignment = str(u.assignment)
	  if uassignment:
	    if not re.search("robClk=0",uassignment):
	      uassignment += ",robClk=0"
	  else:
	    uassignment = "robClk=0"
	  found = False
	  if u.target == bad_loc:
	    for t in new_transitions: #we search for equivalent transitions, which happen specifically if we apply the transformations: consistencyGame and robustTGA successively.  
	      if (t.source == alpha_loc or t.source == beta_loc) and \
		  t.target == bad_loc and \
		  str(t.guard) == uguard and \
		  str(t.assignment) == uassignment:
		found = True
		break
	  if not found:  
	    new_transitions.append(pyuppaal.Transition(source=alpha_loc, \
		    target = u.target, \
		    controllable = False, \
		    guard = uguard, \
		    synchronisation = str(u.synchronisation), \
		    assignment=uassignment))
	    
	    new_transitions.append(pyuppaal.Transition(source=beta_loc, \
		    target = u.target, \
		    controllable = False, \
		    guard = uguard, \
		    synchronisation = str(u.synchronisation), \
		    assignment=uassignment))
    else:
      if t.target != t.source:
	if str(t.assignment):
	  t.assignment.value += ", robClk=0"
	else:
	  t.assignment.value = "robClk=0"
      
  rob_game.transitions.extend(new_transitions)
  for t in removed_transitions:
    rob_game.transitions.remove(t)
     
  return rob_game
  

def consistencyGame(template,delta=0):
  """Prepare a specification for solving the consistency game.
  - indepenent progress condition: 
    - Remove invariants and replace then by input edges whose target is an inconsistent location.
    - Restrict the guards by the invariants: on outcome edges, and on income edges if the clock is not reset
    (invariants are assumed to be only greater bound; clocks are only reset to zero).
  - delta latency of inputs (if delta != 0):
    - increased guard on inputs by delta
    - add edges to inconsistent location for each pair of edges that start from the same location, with the same input synchronisation
    - add as a guard to these edges the conjunction of the two increased guards.
  - Check that output actions are controllable and input non controllable.
  """
  cons_game = copy.deepcopy(template)
  cons_game.name += "ConsTG"
  
  revert_dbmpyuppaal(cons_game)

  incons_loc = pyuppaal.Location(name="INCONS")
  cons_game.locations.append(incons_loc)
  
  if delta:
    for t in cons_game.transitions:
      if str(t.synchronisation)[-1] == '?' and t.target.name.value != "UNIV":
	guardl = str(t.guard).split("&&")
	for i in range(len(guardl)):
	  guardl[i] = enlargedConstraint(guardl[i],delta)
	t.guard.value = ' && '.join(guardl)
    for l in cons_game.locations:
      i=0
      length = len(cons_game.transitions)
      for ti in range(length):
	ti = cons_game.transitions[i]
	if ti.source == l and  ti.target.name.value != "UNIV" and  str(ti.synchronisation)[-1] == '?':
	  for j in range(i+1,length):
	    tj = cons_game.transitions[j]
	    if tj.source == l and  tj.target.name.value != "UNIV" and str(tj.synchronisation) == str(ti.synchronisation):
	      cons_game.transitions.append(pyuppaal.Transition(source=l, \
		  target = incons_loc, \
		  controllable = False, \
		  synchronisation = str(tj.synchronisation), \
		  guard = str(ti.guard) + " && " + str(tj.guard)))
	    j+=1
	i+=1

  for t in cons_game.transitions:
    if str(t.source.invariant):
      if str(t.guard):
	t.guard.value += " && " + str(t.source.invariant)
      else:
	t.guard.value = str(t.source.invariant)
    if t.source != t.target and str(t.target.invariant):
      updatedClk = []
      guardl = str(t.guard).split("&&")
      if guardl[0]=="":
	guardl=[]
      if str(t.assignment):
	updatel = str(t.assignment).split(",")
	for ui in updatel:
	  r = re.search('^\s*([a-zA-Z_]\w*)', ui)
	  updatedClk.append(r.group(1))
      invl = str(t.target.invariant).split("&&")
      for invi in invl:
	r = re.search('([a-zA-Z_]\w*)', invi)
	invClk = r.group(1)
	if not invClk in updatedClk:
	  guardl.append(invi)
      t.guard.value = ' && '.join(guardl)
    if str(t.synchronisation)[-1] == '!':
      t.controllable = True
    else:
      t.controllable = False
    
  for l in cons_game.locations:
    if  str(l.invariant):
      invl = str(l.invariant).split("&&")
      for invi in invl:
	  cons_game.transitions.append(pyuppaal.Transition(source=l, \
		target = incons_loc, \
		controllable = False, \
		guard = contraryConstraint(invi)))
      l.invariant.value = ""
    
  return cons_game
  

def usefulnessGame(template,bad_loc_list=[]):
  """Prepare a specification for solving the usefulness game.
  - Remove invariants and add it to the guards of incoming and outcoming edges.
    If invariants are overpassed the environnement wins. Thus, contrary to the consistency game the spoiler (output player) is force to play.
  - Check that output actions are non controllable and input controllable.
  - all bad locations in bad_loc_list are renamed UNIV, such that there remains only one objective
  - remove self loop on inputs
  """
  useful_game = copy.deepcopy(template)
  useful_game.name += "UsefulTG"
  
  revert_dbmpyuppaal(useful_game)
  
  #removed_transitions = []
  for t in useful_game.transitions:
    #if str(t.synchronisation)[-1] == '?' and t.source == t.target and str(t.assignment) == "":
      #removed_transitions.append(t)
      #continue
    if str(t.source.invariant):
      if str(t.guard):
	t.guard.value += " && " + str(t.source.invariant)
      else:
	t.guard.value = str(t.source.invariant)
    if t.source != t.target and str(t.target.invariant):
      updatedClk = []
      guardl = str(t.guard).split("&&")
      if guardl[0]=="":
	guardl=[]
      if str(t.assignment):
	updatel = str(t.assignment).split(",")
	for ui in updatel:
	  r = re.search('^\s*([a-zA-Z_]\w*)', ui)
	  updatedClk.append(r.group(1))
      invl = str(t.target.invariant).split("&&")
      for invi in invl:
	r = re.search('([a-zA-Z_]\w*)', invi)
	invClk = r.group(1)
	if not invClk in updatedClk:
	  guardl.append(invi)
      t.guard.value = ' && '.join(guardl)
    if 	str(t.synchronisation)[-1] == '!':
      t.controllable = False
    else:
      t.controllable = True
   
  for l in useful_game.locations:
    if l.name.value in bad_loc_list:
      l.name.value =  "UNIV"
    if str(l.invariant):
      l.invariant.value = ""
  
  #for t in removed_transitions:
    #useful_game.transitions.remove(t)
    
  return useful_game  

  
def revert_dbmpyuppaal(template):
  """Revert the modifications made by dbmpyuppaal to the guard and invariant of the template"""
  for t in template.transitions:
    if hasattr(t,"guard_str"):
      t.guard =  pyuppaal.Label("guard", t.guard_str)
      del t.guard_str
    if t.synchronisation and not isinstance(t.synchronisation,pyuppaal.Label):
      t.synchronisation =  pyuppaal.Label("synchronisation",str(t.synchronisation))
  for l in template.locations:
    if isinstance(l.name,str):
      l.name = pyuppaal.Label("name", l.name)
    if hasattr(l,"invariant_str"):
      l.invariant =  pyuppaal.Label("invariant", l.invariant_str)
      del l.invariant_str

        
def conjunction(spec1,spec2,outputs,inputs,lu_name="UNIV"):
  """Conjunction of two specifications sharing the same alphabet.
  We assume the two specifications have no clock in common (or with the same name)."""
  revert_dbmpyuppaal(spec1)
  revert_dbmpyuppaal(spec2)

  conj = pyuppaal.Template(spec1.name+"AND"+spec2.name, \
	  declaration=spec1.declaration+"\n"+spec2.declaration)
  
  loc_dict = {}
  lu = None
  for l1 in spec1.locations:
    for l2 in spec2.locations:
      if str(l1.name) == lu_name or str(l2.name) == lu_name:
	if not lu: # if multiple we merge lu locations
	  lu = pyuppaal.Location(name="UNIV")
	  conj.locations.append(lu)
	  continue;
      else:
	l1_2 = pyuppaal.Location(name=str(l1.name)+str(l2.name))
	conj.locations.append(l1_2)
	loc_dict[(l1,l2)]=l1_2
	if str(l1.invariant):
	  if str(l2.invariant):
	    l1_2.invariant = pyuppaal.Label("invariant", str(l1.invariant)+' && '+str(l2.invariant))
	  else:
	    l1_2.invariant = pyuppaal.Label("invariant", str(l1.invariant))
	elif str(l2.invariant):
	  l1_2.invariant = pyuppaal.Label("invariant", str(l2.invariant))
	if spec1.initlocation == l1 and spec2.initlocation == l2:
	  conj.initlocation = l1_2
	  
  for t1 in spec1.transitions:
    if str(t1.source.name) != lu_name:
      for t2 in spec2.transitions:
	if str(t2.source.name) != lu_name:
	  if str(t1.synchronisation) == str(t2.synchronisation): # synchronisations, including channel type
	    assert(t1.controllable == t2.controllable)
	    if str(t1.target.name) == lu_name or str(t2.target.name) == lu_name:
	      target = lu
	    else:
	      target = loc_dict[(t1.target,t2.target)]
	    t1_2 = pyuppaal.Transition(source=loc_dict[(t1.source,t2.source)], \
		      target = target, \
		      controllable = t1.controllable, \
		      synchronisation = str(t1.synchronisation))
	    if str(t1.guard):
	      if str(t2.guard):
		t1_2.guard = pyuppaal.Label("guard", str(t1.guard)+' && '+str(t2.guard))
	      else:
		t1_2.guard = pyuppaal.Label("guard", str(t1.guard))
	    elif str(t2.guard):
	      t1_2.guard = pyuppaal.Label("guard", str(t2.guard))
	    if str(t1.assignment):
	      if str(t2.assignment):
		t1_2.assignment = pyuppaal.Label("assignment", str(t1.assignment)+', '+str(t2.assignment))
	      else:
		t1_2.assignment = pyuppaal.Label("assignment", str(t1.assignment))
	    elif str(t2.assignment):
	      t1_2.assignment = pyuppaal.Label("assignment", str(t2.assignment))
	    conj.transitions.append(t1_2)
  
  #UNIV location
  if lu:
    for action in outputs:
      conj.transitions.append(pyuppaal.Transition(source=lu, \
		target = lu, \
		controllable = False, \
		synchronisation = action+"!"))
    for action in inputs:
      conj.transitions.append(pyuppaal.Transition(source=lu, \
	      target = lu, \
	      controllable = True, \
	      synchronisation = action+"?"))
	
  return conj
  
      
def composition(spec1,spec2,outputs1,outputs2,inputs1,inputs2):
  """Parallel composition of two specifications
    We assume the two specifications have no clock in common (or with the same name)."""
  revert_dbmpyuppaal(spec1)  
  revert_dbmpyuppaal(spec2)

  compo = pyuppaal.Template(spec1.name+"PAR"+spec2.name, \
	  declaration=spec1.declaration+"\n"+spec2.declaration)
  
  loc_dict = {}
  lu = None
  for l1 in spec1.locations:
    for l2 in spec2.locations:
      if str(l1.name) == "UNIV" or str(l2.name) == "UNIV":
	if not lu:
	  lu = pyuppaal.Location(name="UNIV")
	  compo.locations.append(lu)
	  continue;
      else:
	l1_2 = pyuppaal.Location(name=str(l1.name)+"PAR"+str(l2.name))
	loc_dict[(l1,l2)]=l1_2
	compo.locations.append(l1_2)
	if str(l1.invariant):
	  if str(l2.invariant):
	    l1_2.invariant = pyuppaal.Label("invariant", str(l1.invariant)+' && '+str(l2.invariant))
	  else:
	    l1_2.invariant = pyuppaal.Label("invariant", str(l1.invariant))
	elif str(l2.invariant):
	  l1_2.invariant = pyuppaal.Label("invariant", str(l2.invariant))
	if spec1.initlocation == l1 and spec2.initlocation == l2:
	  compo.initlocation = l1_2
	
  for t1 in spec1.transitions:
    if str(t1.source.name) != "UNIV":
      if str(t1.synchronisation)[:-1] not in outputs2+inputs2:			 #t1 is not synchronized
	for l2 in spec2.locations:
	  if str(l2.name) != "UNIV":
	    if str(t1.target.name) != "UNIV":
	      target = loc_dict[(t1.target,l2)]
	    else:
	      target = lu
	    t1_2 = pyuppaal.Transition(source=loc_dict[(t1.source,l2)], \
		      target = target, \
		      controllable = t1.controllable, \
		      guard = str(t1.guard), \
		      synchronisation = str(t1.synchronisation), \
		      assignment = str(t1.assignment))
	    compo.transitions.append(t1_2)
      else: #t1 is synchronized
	for t2 in spec2.transitions:
	  if str(t2.source.name) != "UNIV":
	    if str(t1.synchronisation)[:-1] == str(t2.synchronisation)[:-1]:	#same channel
	      if str(t1.target.name) == "UNIV" or str(t2.target.name) == "UNIV":
		target = lu
	      else:
		target = loc_dict[(t1.target,t2.target)]
	      t1_2 = pyuppaal.Transition(source=loc_dict[(t1.source,t2.source)], \
			target = target, \
			controllable = t1.controllable and t2.controllable)
	      if str(t1.guard):
		if str(t2.guard):
		  t1_2.guard = pyuppaal.Label("guard", str(t1.guard)+' && '+str(t2.guard))
		else:
		  t1_2.guard = pyuppaal.Label("guard", str(t1.guard))
	      elif str(t2.guard):
		t1_2.guard = pyuppaal.Label("guard", str(t2.guard))	  
	      if str(t1.assignment):
		if str(t2.assignment):
		  t1_2.assignment = pyuppaal.Label("assignment", str(t1.assignment)+', '+str(t2.assignment))
		else:
		  t1_2.assignment = pyuppaal.Label("assignment", str(t1.assignment))
	      elif str(t2.assignment):
		t1_2.assignment = pyuppaal.Label("assignment", str(t2.assignment))
	      if str(t1.synchronisation) ==  str(t2.synchronisation): 		#input-input synchronisation
		t1_2.synchronisation = pyuppaal.Label("synchronisation", str(t1.synchronisation))
	      else: 								#input-output synchronisation
		t1_2.synchronisation = pyuppaal.Label("synchronisation", str(t1.synchronisation)[:-1]+"!")
	      compo.transitions.append(t1_2)
		
  for t2 in spec2.transitions:
    if str(t2.source.name) != "UNIV":
      if str(t2.synchronisation)[:-1] not in outputs1+inputs1: 			#t2 is not synchronized
	for l1 in spec1.locations:
	  if str(l1.name) != "UNIV":
	    if str(t2.target.name) != "UNIV":
	      target = loc_dict[(l1,t2.target)]
	    else:
	      target = lu
	    t1_2 = pyuppaal.Transition(source=loc_dict[(l1,t2.source)], \
		      target = target, \
		      controllable = t2.controllable, \
		      guard = str(t2.guard), \
		      synchronisation = str(t2.synchronisation), \
		      assignment = str(t2.assignment))
	    compo.transitions.append(t1_2)

  #UNIV location
  if lu:
    for action in outputs1+outputs2:
      compo.transitions.append(pyuppaal.Transition(source=lu, \
		target = lu, \
		controllable = False, \
		synchronisation = action+"!"))
    for action in inputs1:
      if action not in outputs1+outputs2:
	compo.transitions.append(pyuppaal.Transition(source=lu, \
		target = lu, \
		controllable = True, \
		synchronisation = action+"?"))
    for action in inputs2:
      if action not in outputs1+outputs2 and action not in inputs1:
	compo.transitions.append(pyuppaal.Transition(source=lu, \
		target = lu, \
		controllable = True, \
		synchronisation = action+"?"))	

  return compo
     
     

def invert(spec,outputs):
  """Invert a specification such that:
   - input transitions become output
   - output transitions become input
   - invariants are removed and added to the guards of incident edges
   - if the invariant is violated a transition is taken to an error location UNSAFE
   - the model is made input enabled: any input not accepted previously is accepted by a transition to  the location UNSAFE
  """
  invert_spec = copy.deepcopy(spec)
  invert_spec.name += "Invert"
  revert_dbmpyuppaal(invert_spec)

  unsafe_loc = pyuppaal.Location(name="UNSAFE")
  invert_spec.locations.append(unsafe_loc)
  
  for t in invert_spec.transitions:
    if t.controllable: # input
      t.synchronisation.value = t.synchronisation.value[:-1] + '!'
      t.controllable = False
    else: # output
      t.synchronisation.value = t.synchronisation.value[:-1] + '?'
      t.controllable = True
      
    if str(t.source.invariant):
      if str(t.guard):
	t.guard.value += " && " + str(t.source.invariant)
      else:
	t.guard.value = str(t.source.invariant)
    
  for l in invert_spec.locations:
    if  str(l.invariant):
      invl = str(l.invariant).split("&&")
      for invi in invl:
	  invert_spec.transitions.append(pyuppaal.Transition(source=l, \
		target = unsafe_loc, \
		controllable = False, \
		guard = contraryConstraint(invi)))
      l.invariant.value = ""
      
  makeInputEnabled(invert_spec,outputs,unsafe_loc)
    
  return invert_spec
  

def makeInputEnabled(spec,inputs,error_loc):
    """ Make a specification input enabled by linking all non accepted inputs to the error location"""
    
    for l in spec.locations:
      for channel in inputs:
	conj_guards = []
	skip = False
	for t in spec.transitions:
	  if t.source == l and str(t.synchronisation)[:-1] == channel:
	    if not str(t.guard):
	      skip = True
	      break
	    else:
	      contrary_guard = []
	      guardl = str(t.guard).split("&&")
	      for guardi in guardl:
		contrary_guard.append(contraryConstraint(guardi))
	      conj_guards.append(contrary_guard)
	
	if not skip:
	  disj_guards= [[]]
	  for disj in conj_guards:
	    new_disj_guards = []
	    for conj in disj_guards:
	      for i in range(len(disj)-1):
		new_conj = copy.copy(conj)
		new_conj.append(disj[i])
		new_disj_guards.append(new_conj)
	      conj.append(disj[len(disj)-1])
	    disj_guards += new_disj_guards
	
	  for conj in disj_guards:
	    spec.transitions.append(pyuppaal.Transition(source=l, \
		target = error_loc, \
		controllable = True, \
		guard = ' && '.join(conj), \
		synchronisation = channel+"?"))
		
    
    
    
    
    
    
    
    