"""PyECDAR, a tool for timed games and timed specifications.
Copyright (C) INRIA Rennes - Louis-Marie Traonouez

Louis-Marie.Traonouez@inria.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

from fractions import Fraction
import model_transformations
import dbmpyuppaal
import pyuppaal

import pyrobustcegar
import pytiga
import pyreachability

import logging

logging.basicConfig()
logger = logging.getLogger('ECDAR')

logger.setLevel(logging.INFO)
pyrobustcegar.logger.setLevel(logging.INFO)
pytiga.logger.setLevel(logging.INFO)
pyreachability.logger.setLevel(logging.INFO)
dbmpyuppaal.dbmpyuppaal.my_logger.setLevel(logging.ERROR)

def loadModel(filename):
  try:
    nta = pyuppaal.NTA.from_xml(filename)
  except IOError as (errno, strerror):
    logger.error("I/O error({0}): {1}".format(errno, strerror))
  else:
    return Workspace(nta,filename)


class Workspace:
  def __init__(self,nta,filename):
    self.nta=nta
    self.filename = filename
    
  def addTemplate(self,template_model):
    """template_model is either a game or a specification"""
    model_transformations.revert_dbmpyuppaal(template_model.template)
    template_model.template.layout(auto_nails=True)
    self.nta.templates.append(template_model.template)
    
  def save(self,filename=""):
    for template in self.nta.templates:
      model_transformations.revert_dbmpyuppaal(template)

    if not filename:
      filename = self.filename
    try:
      f = open(filename, 'w')
      f.write(self.nta.to_xml())
      f.close()
    except IOError as (errno, strerror):
      logger.error("I/O error({0}): {1}".format(errno, strerror))

  def getSpecification(self,name):
    ts = [t for t in self.nta.templates if t.name == name] 
    if len(ts) != 1:
      logger.error("Key error with template name \""+name+"\"")
      return None
    else:
      return Specification(ts[0],self)
  
  def getTimedGame(self,name,objective):
    ts = [t for t in self.nta.templates if t.name == name] 
    if len(ts) != 1:
      logger.error("Key error with template name \""+name+"\"")
      return None
    else:
      return TimedGame(ts[0],objective,self)
   
   
class TimedGame:
  def __init__(self,template,objective,workspace):
    self.template = template
    self.objective = objective
    self.workspace = workspace
    model_transformations.revert_dbmpyuppaal(self.template)
    self.dbm_model = False

  def solve(self):
    model_transformations.revert_dbmpyuppaal(self.template)
    dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[self.template])
    game = pytiga.SafetyTimedGame(self.workspace.nta,self.template,self.objective)
    return game.solve()
  
  def scaledTemplate(self,factor):
    return TimedGame(model_transformations.scaledTemplateConstants(self.template,factor),self.objective,self)

    
class Specification:
  def __init__(self,template,workspace):
    self.template = template
    self.workspace = workspace
    self.inputs=[]
    self.outputs=[]

    for transition in template.transitions:
      if str(transition.synchronisation):
	sync_type = str(transition.synchronisation)[-1]
	channel_name = str(transition.synchronisation)[:-1]
	if sync_type is '!':
	  if channel_name not in self.outputs:
	    self.outputs.append(channel_name)
	elif sync_type is '?':
	  if channel_name not in self.inputs:
	    self.inputs.append(channel_name)
	else:
	  logger.warning("unknown synchronisation type: " + sync_type)
  
  def isSpecification(self):
    """Check if the template syntax correspond to a specification:
    - all transition are synchronized
    - input enabled
    - disjoint input and output
    - invariants are upper bounds
    - rectangular guards and invariants (not check: should yield an error when parsing through dbmpyuppaal)
    - reset values are zero
    - all the clocks are local
    """
    model_transformations.revert_dbmpyuppaal(self.template)
    dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[self.template])
    if len(self.workspace.nta.clocks) > 0:
      logger.error("the model contains global clocks.")
      return False
    for transition in self.template.transitions:
      if not str(transition.synchronisation):
	logger.error("transition \"" + str(transition) + "\" has no synchronisation.")
	return false
      if str(transition.synchronisation)[:-1] in self.inputs and str(transition.synchronisation)[:-1] in self.outputs:
	logger.error("inputs and outputs channel are not disjoint.")
	return false
      for (clock,value) in transition.update:
	if value != 0:
	  logger.error("transition \"" + str(transition) + "\" has a clock reset to a non zero value.")
	  return False
    for location in self.template.locations:
      for i in self.inputs:
	found = False
	for transition in self.template.forward[location]:
	  if transition.synchronisation.channel.name == i:
	    found = True
	    break
	if not found:
	  logger.error("location \"" + str(location.name) + "\" is not imput enabled.")
	  return False
      if not location.invariant.hasZero():
	logger.error("location \"" + str(location.name) + "\" has a lower bound invariant.")
	return False
    return True
      

  #def isImplementation(self):
    #"""Check if the template syntax correspond to an implementation automata:
    #- output urgency (except for the universal location)
    #- independent progress
    #"""
    #if not self.isSpecification():
      #return False
    
  
  def implemPerturbation(self,delta="delta",param=0):
    """Construct the delta-perturbation of an implementation.
      - delta is either the value of delta, or the name of the parameter (by default: delta)
      - param is the value of delta as a parameter (by default 0)
      Note that if delta is a value, param is useless.
      Return a new template corresponding, and if delta is a parameter, add a const declaration in the nta.
      """
    if isinstance(delta,str):
      if self.workspace.nta.declaration.find(delta) < 0:
	self.workspace.nta.declaration += "\n const int " + delta + " = " + str(param) +";"
      else:
	logger.warning("parameter " + delta + " already defined.")
      return Specification(model_transformations.implemPerturbation(self.template,delta),self.workspace)
    else:
      try:
	fdelta = Fraction(str(delta))
      except:
	logger.error("invalid input parameter: "+ str(delta))
	raise
      scaled_impl = model_transformations.scaledTemplateConstants(self.template,fdelta.denominator)
      return Specification(model_transformations.implemPerturbation(scaled_impl,fdelta.numerator),self.workspace)
  
  def scaledTemplate(self,factor):
    return Specification(model_transformations.scaledTemplateConstants(self.template,factor),self.workspace)
  
  
  ### Consistency checking  ----------------------------------------------------------------------------------------
  
  def getConsistencyGame(self):
    return TimedGame(model_transformations.consistencyGame(self.template),"INCONS",self.workspace)
    
  def getRobustConsistencyGame(self,delta="delta",param=0):
    """Construct the robust consistency game. 
      - delta is either the value of delta, or the name of the parameter (by default: delta)
      - param is the value of delta as a parameter (by default 0)
      Note that if delta is a value, param is useless.
      Return the template corresponding to the game and if delta is a parameter, add a const declaration in the nta.
      """
      
    if isinstance(delta,str):
      cons_game = model_transformations.consistencyGame(self.template,delta)
      if self.workspace.nta.declaration.find(delta) < 0:
	self.workspace.nta.declaration += "\n const int " + delta + " = " + str(param) +";"
      return TimedGame(model_transformations.robustTGA(cons_game,delta,"INCONS"),"INCONS",self.workspace)
    else:
      try:
	fdelta = Fraction(str(delta))
      except:
	logger.error("invalid input parameter: "+ str(delta))
	raise
      scaled_game = model_transformations.scaledTemplateConstants(self.template,fdelta.denominator)
      cons_game = model_transformations.consistencyGame(scaled_game,fdelta.numerator)
      return TimedGame(model_transformations.robustTGA(cons_game,fdelta.numerator,"INCONS"),"INCONS",self.workspace)
  
    
  def isConsistent(self):
    cons_game = model_transformations.consistencyGame(self.template)
    dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[cons_game])
    
    timeGame = pytiga.SafetyTimedGame(self.workspace.nta,cons_game,"INCONS")
    return timeGame.solve()
  
  def isRobustConsistent(self,delta):
    """Delta is a decimal value"""
    try:
      fdelta = Fraction(str(delta))
    except:
      logger.error("invalid input parameter: "+ str(delta))
      raise
    scaled_game = model_transformations.scaledTemplateConstants(self.template,fdelta.denominator)
    cons_game = model_transformations.consistencyGame(scaled_game,fdelta.numerator)
    robust_cons_game = model_transformations.robustTGA(cons_game,fdelta.numerator,"INCONS")

    dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[robust_cons_game])
    
    timeGame = pytiga.SafetyTimedGame(self.workspace.nta,robust_cons_game,"INCONS")
    return timeGame.solve()
  
    
  def maxRobustConsistent(self,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    cons_game = model_transformations.consistencyGame(self.template,"delta")
    robust_cons_game = model_transformations.robustTGA(cons_game,"delta","INCONS")
    if self.workspace.nta.declaration.find("delta") < 0:
      self.workspace.nta.declaration += "\n const int delta = " + str(delta_max) +";"
    else:
      logger.warning("parameter delta already defined.")
    
    cegarAlgo = pyrobustcegar.CegarForRobustness(self.workspace.nta,robust_cons_game,"INCONS",2,str(confidence),delta_max)
    return cegarAlgo.cegar()
    
  def maxRobustConsistentBS(self,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    cons_game = model_transformations.consistencyGame(self.template,"delta")
    robust_cons_game = model_transformations.robustTGA(cons_game,"delta","INCONS")
    if self.workspace.nta.declaration.find("delta") < 0:
      self.workspace.nta.declaration += "\n const int delta = " + str(delta_max) +";"
    else:
      logger.warning("parameter delta already defined.")
    
    cegarAlgo = pyrobustcegar.CegarForRobustness(self.workspace.nta,robust_cons_game,"INCONS",2,str(confidence),delta_max)
    return cegarAlgo.binary_refine()
    
    
  ### Usefulness checking --------------------------------------------------------------------------------------------------
  
  def getUsefulnessGame(self,bad_loc_list=[]):
    return TimedGame(model_transformations.usefulnessGame(self.template,bad_loc_list=[]),"UNIV",self.workspace)
    
  def getRobustUsefulnessGame(self,delta="delta",param=0,bad_loc_list=[]):
    if isinstance(delta,str):
      useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list=[])
      if self.workspace.nta.declaration.find(delta) < 0:
	self.workspace.nta.declaration += "\n const int " + delta + " = " + str(param) +";"
      return TimedGame(model_transformations.robustTGA(useful_game,delta,"UNIV"),"UNIV",self.workspace)
    else:
      try:
	fdelta = Fraction(str(delta))
      except:
	logger.error("invalid input parameter: "+ str(delta))
	raise
      useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
      scaled_game = model_transformations.scaledTemplateConstants(useful_game,fdelta.denominator)
      return TimedGame(model_transformations.robustTGA(scaled_game,fdelta.numerator,"UNIV"),"UNIV",self.workspace)
    
  def isUseful(self,bad_loc_list=[]):
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[useful_game])
    
    timeGame = pytiga.SafetyTimedGame(self.workspace.nta,useful_game,"UNIV")
    return timeGame.solve()
    
  def isRobustUseful(self,delta,bad_loc_list=[]):
    """Delta is a decimal value"""
    try:
      fdelta = Fraction(str(delta))
    except:
      logger.error("invalid input parameter: "+ str(delta))
      raise    
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    scaled_game = model_transformations.scaledTemplateConstants(useful_game,fdelta.denominator)
    robust_useful_game = model_transformations.robustTGA(scaled_game,fdelta.numerator,"UNIV")

    dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[robust_useful_game])
    
    timeGame = pytiga.SafetyTimedGame(self.workspace.nta,robust_useful_game,"UNIV")
    return timeGame.solve()
    
  def maxRobustUseful(self,delta_max,confidence,bad_loc_list=[]):
    """ delta_max is an integer
        confidence is a decimal"""
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    robust_useful_game = model_transformations.robustTGA(useful_game,"delta","UNIV")
    if self.workspace.nta.declaration.find("delta") < 0:
      self.workspace.nta.declaration += "\n const int delta = " + str(delta_max) +";"
    else:
      logger.warning("parameter delta already defined.")
    
    cegarAlgo = pyrobustcegar.CegarForRobustness(self.workspace.nta,robust_useful_game,"UNIV",2,str(confidence),delta_max)
    return cegarAlgo.cegar()
    
  def maxRobustUsefulBS(self,delta_max,confidence,bad_loc_list=[]):
    """ delta_max is an integer
        confidence is a decimal"""
    useful_game = model_transformations.usefulnessGame(self.template,bad_loc_list)
    robust_useful_game = model_transformations.robustTGA(useful_game,"delta","UNIV")
    if self.workspace.nta.declaration.find("delta") < 0:
      self.workspace.nta.declaration += "\n const int delta = " + str(delta_max) +";"
    else:
      logger.warning("parameter delta already defined.")
    
    cegarAlgo = pyrobustcegar.CegarForRobustness(self.workspace.nta,robust_useful_game,"UNIV",2,str(confidence),delta_max)
    return cegarAlgo.binary_refine()
    
  ### Refinement -------------------------------------------------------------------------------------------------------------------------
  
  def refine(self,big_spec):
    model_transformations.revert_dbmpyuppaal(self.template)
    model_template = model_transformations.composition(self.template, \
			    model_transformations.invert(big_spec.template,big_spec.outputs), \
			    self.outputs, \
			    big_spec.inputs, \
			    self.inputs, \
			    big_spec.outputs)
    
    dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[model_template])

    refCheck = pyreachability.ReachabilityAnalysis(self.workspace.nta,model_template,"UNSAFE",True)
    return refCheck.solve()
    
  #def refine(self,big_spec):
    #model_transformations.revert_dbmpyuppaal(self.template)
    #model_transformations.revert_dbmpyuppaal(big_spec.template)
    
    #dbmpyuppaal.parse_nta_dbm(self.workspace.nta)
    #dbmpyuppaal.parse_templates_dbm(self.workspace.nta,[self.template,big_spec.template])

    #refGame = pytigarefinement.RefinementTimedGame(self.workspace.nta,self.template,big_spec.template,False)
    #return refGame.solve()  
       
  def maxRefine(self,big_spec,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    perturbed = self.implemPerturbation("1*delta")
    model_template = model_transformations.composition(perturbed.template, \
			    model_transformations.invert(big_spec.template,big_spec.outputs), \
			    self.outputs, \
			    big_spec.inputs, \
			    self.inputs, \
			    big_spec.outputs)
    if self.workspace.nta.declaration.find("delta") < 0:
      self.workspace.nta.declaration += "\n const int delta = " + str(delta_max) +";"
    else:
      logger.warning("parameter delta already defined.")
    
    cegarAlgo = pyrobustcegar.CegarForRobustness(self.workspace.nta,model_template,"UNSAFE",1,str(confidence),delta_max)
    return cegarAlgo.cegar()
    
  def maxRefineBS(self,big_spec,delta_max,confidence):
    """ delta_max is an integer
        confidence is a decimal"""
    perturbed = self.implemPerturbation("delta")
    model_template = model_transformations.composition(perturbed.template, \
			    model_transformations.invert(big_spec.template,big_spec.outputs), \
			    self.outputs, \
			    big_spec.inputs, \
			    self.inputs, \
			    big_spec.outputs)
    if self.workspace.nta.declaration.find("delta") < 0:
      self.workspace.nta.declaration += "\n const int delta = " + str(delta_max) +";"
    else:
      logger.warning("parameter delta already defined.")
    
    cegarAlgo = pyrobustcegar.CegarForRobustness(self.workspace.nta,model_template,"UNSAFE",1,str(confidence),delta_max)
    return cegarAlgo.binary_refine()
    
  def __le__(self,big_spec):
    return self.refine(big_spec)
    
  def __ge__(self,small_spec):
    return small_spec.refine(self)
    
    
  ### Conjunction and composition -------------------------------------------------------------------------------------------------------------------------

  def sameAlphabet(self,spec2):
    """ Check if two specifications have the same action alphabet.
	This condition is required to compute the conjunction.  """
    for a in self.inputs:
      if a not in spec2.inputs:
	return False
    for a in self.outputs:
      if a not in spec2.outputs:
	return False
    return True
    
  def conjunction(self,spec2,lu_name="UNIV"):
    if not self.sameAlphabet(spec2):
      logger.error("the specifications have two different alphabet.")
      return None
    if self.workspace != spec2.workspace:
      logger.warning("the specifications belong to different workspaces.")
    return Specification(model_transformations.conjunction(self.template,spec2.template,self.outputs,self.inputs,lu_name),self.workspace)
    
  def composable(self,spec2):
    """Check that the two specifications are composable: their output alphabets are disjoint"""
    for a in self.outputs:
      if a in spec2.outputs:
	return False
    return True
  
  def composition(self,spec2):
    if not self.composable(spec2):
      logger.error("the specifications are not composable.")
      return None
    if self.workspace != spec2.workspace:
      logger.warning("the specifications belong to different workspaces.")
    return Specification(model_transformations.composition(self.template,spec2.template,self.outputs,spec2.outputs,self.inputs,spec2.inputs),self.workspace)

  def __and__(self,spec2):
    return self.conjunction(spec2)
    
  def __or__(self,spec2):
    return self.composition(spec2)
  

if __name__ == '__main__':
  import sys

  W = loadModel(sys.argv[1])
  S=W.getSpecification(sys.argv[2])
  T=W.getSpecification(sys.argv[3])
  delta=sys.argv[4]
  
  G = S.getRobustConsistencyGame("delta",10)
  W.addTemplate(G)
  W.save(filename="coffee_machine-out.xml")
  
  print T <= S
  print S <= T
  print S.isConsistent()

  print T.isRobustConsistent(delta)
  print S.maxRobustConsistent(10,0.1)

  

