"""PyECDAR, a tool for timed games and timed specifications.
Copyright (C) INRIA Rennes - Louis-Marie Traonouez

Louis-Marie.Traonouez@inria.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import sys
import logging

import pydbm
import pyparamexpr
import pyppl

logging.basicConfig()
logger = logging.getLogger('REACH')


class DiscreteTransition:
    def __init__(self, source, destination, model_transition):
        assert(isinstance(source,pydbm.Federation))
        assert(isinstance(destination,pydbm.Federation))
        
        self.source = source
        self.destination = destination
        self.model_transition = model_transition
    
    def __str__(self):
        return str(self.source) + '--->>>' + str(self.destination) + '[' + str(self.model_transition) + ']'
    def __hash__(self):
        return hash((self.source,self.destination,self.model_transition))
    def __eq__(self, arg2):
        return  (self.source == arg2.source) and \
                (self.destination == arg2.destination) and \
                (self.model_transition == arg2.model_transition)

class ReachabilityAnalysis:    
    def __init__(self, model, goal_template, goal_location_name, safety):
        self.model = model 
        self.goal_template = goal_template
        self.goal_location_name = goal_location_name
        self.context = model.context
        self.safety = safety
	for l in self.goal_template.locations:
	  l.zone = self.context.getEmptyFederation()
	  locations = l.name.split("PAR")
	  l.accepting = (goal_location_name in locations)
	  l.history = []
	self.counterexample = []
    def takeTransition(self, zone, model_transition):
        ret_zone = zone.copy()
        ret_zone &= model_transition.guard
        if ret_zone.isEmpty():
            return None
        for (clock, value) in model_transition.update:
	  if value != 0:
            logger.error("Clock reset to non-zero value")
            sys.exit(1)
	  else:
            ret_zone.resetValueInPlace(clock)
        ret_location = model_transition.target
        ret_zone &= ret_location.invariant
	if ret_zone.isEmpty():
            return None

        ret_zone.upInPlace()
        ret_zone &= ret_location.invariant
        ret_zone.extrapolateMaxBoundsInPlace(self.model.clock_bounds)

        transition = DiscreteTransition(zone,ret_zone,model_transition)
        return transition
    def next(self, loc):
        ret = []
        for transition in self.goal_template.forward[loc]:       
	  next_discrete_transition = self.takeTransition(loc.zone, transition)
	  if not (next_discrete_transition is None):
	    ret.append(next_discrete_transition)
        return ret
    def solve(self):
        waiting = []
        initial = self.goal_template.initlocation
	initial.zone = self.context.getZeroFederation()
        initial.zone.upInPlace()
        initial.zone &= initial.invariant
        initial.zone.extrapolateMaxBoundsInPlace(self.model.clock_bounds)
        if (initial.name == self.goal_location_name):
	  return not self.safety
	for e_ in self.next(initial):
	  logger.debug("adding to wait next initial: " + str(e_.model_transition.target) + "#" + str(e_.destination))
          waiting.append(e_)
      
        i = 0
        while waiting:
            e = waiting.pop()
            t = e.model_transition
            S = e.source
            loc = t.source
            S_ = e.destination
            loc_ = t.target
            t.target.history = t.source.history + [t]
            i += 1
            if not S_ <= loc_.zone: # not included
	        logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) +' (from ' + str(t) + ')')
                loc_.zone |= S_
                if not loc_.accepting:
                    for e_ in self.next(loc_):
			logger.debug("adding to wait next: " + str(e_.model_transition.target) + "#" + str(e_.destination))
			if not e_ in waiting:
			    waiting.append(e_)
                else:
		    logger.info(str(i) + " states visited.")
		    self.counterexample = t.target.history
		    return not self.safety
	
	logger.info(str(i) + " states visited.")
        return self.safety
        
        
        
    def refine(self,delta_value, delta_name="delta"):
	""" Refine a counterexample run to extract parametrically the minimum value of delta.
	"""
	self.param_context = pyppl.ParamContext(self.context.clocks,delta_name)
	self.initial_domain = self.param_context.getInitialDomain(0,delta_value)
	logger.debug("initial domain: " + str(self.initial_domain))
	
	if not self.counterexample :
	  logger.debug("no counterexample computed")
	  return self.param_context.getEmptyParamFederation()
	
	param_zone = self.param_context.getZeroDomain(0,delta_value)
	param_zone.upInPlace()
	param_zone &= pyparamexpr.parse_param_invariant_or_guard( self.goal_template.initlocation.invariant_str,delta_name,self.param_context)
	param_zone &= self.initial_domain
      
	for t in self.counterexample:	  
	  guard = pyparamexpr.parse_param_invariant_or_guard(t.guard_str,delta_name,self.param_context)
	  update = [self.param_context.getIdVariable(clock.name.split('.')[1]) for (clock,value) in t.update]
	  
	  param_zone.postInPlace(guard,update)
	  param_zone.upInPlace()
	  param_zone &=  pyparamexpr.parse_param_invariant_or_guard(t.target.invariant_str,delta_name,self.param_context)
	  param_zone &= self.initial_domain
	  logger.debug("current domain: " + str(param_zone))

	print param_zone.minimizeParam()	
	return param_zone.minimizeParam()	
	
	      
   
#if __name__ == '__main__':

    #model = pyuppaal.NTA.from_xml(sys.argv[1])
    
    #dbmpyuppaal.parse_nta_dbm(model)
    
    #goal_template = model.getTemplateByName(sys.argv[2].split('.')[0])
    #goal_location_name = sys.argv[2].split('.')[1]
    
    #dbmpyuppaal.parse_templates_dbm(model,[goal_template])

    #game = SafetyTimedGame(model, goal_template, goal_location_name) 
    #print game.solve()
    #game.printDot('game.dot')

    
