"""PyECDAR, a tool for timed games and timed specifications.
Copyright (C) INRIA Rennes - Louis-Marie Traonouez

Louis-Marie.Traonouez@inria.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import sys
import logging

import pydbm
import pyparamexpr
import pyppl

    
logging.basicConfig()
logger = logging.getLogger('TIGA')


class DiscreteTransition:
    def __init__(self, source, destination, model_transition):
        assert(isinstance(source,pydbm.Federation))
        assert(isinstance(destination,pydbm.Federation))
        
        self.source = source
        self.destination = destination
        self.model_transition = model_transition
    def pred(self, zone): # returns preimage of zone
        ret = zone.copy()
        for (clock, value) in self.model_transition.update:
	  if value != 0:
            logger.error("Clock reset to non-zero value")
            sys.exit(1)
	  else:
            ret &= (clock == 0)
            ret.freeClockInPlace(clock)
        ret &= self.model_transition.guard
        ret &= self.source
        return ret
    def __str__(self):
        return str(self.source) + '--->>>' + str(self.destination) + '[' + str(self.model_transition) + ']'
    def __hash__(self):
        return hash((self.source,self.destination,self.model_transition))
    def __eq__(self, arg2):
        return  (self.source == arg2.source) and \
                (self.destination == arg2.destination) and \
                (self.model_transition == arg2.model_transition)

class SafetyTimedGame:    
    def __init__(self, model, goal_template, goal_location_name):
        self.model = model 
        self.goal_template = goal_template
        self.goal_location_name = goal_location_name
        self.context = model.context
	for l in self.goal_template.locations:
	  l.zone = self.context.getEmptyFederation()
	  l.depend = []
	  #l.depend = set()
	  l.accepting = (l.name == self.goal_location_name)
	  l.accepting_zone = self.context.getEmptyFederation()
	  #l.strategy_domain = self.context.getEmptyFederation()
        self._forward = {} 	# dictionnary that store for each state the next reachable states
        self.solved=0 		# the game is initially undetermined
        self.strategy = []	# ordered list of the states refined
    def takeTransition(self, zone, model_transition):
        ret_zone = zone.copy()
        ret_zone &= model_transition.guard
        if ret_zone.isEmpty():
            return None
        for (clock, value) in model_transition.update:
	  if value != 0:
            logger.error("Clock reset to non-zero value")
            sys.exit(1)
	  else:
            ret_zone.resetValueInPlace(clock)
        ret_location = model_transition.target
        ret_zone &= ret_location.invariant
	if ret_zone.isEmpty():
            return None

        ret_zone.upInPlace()
        ret_zone &= ret_location.invariant
        ret_zone.extrapolateMaxBoundsInPlace(self.model.clock_bounds)

        transition = DiscreteTransition(zone,ret_zone,model_transition)
        
	#loc_.depend.add(model_transition)
        if not model_transition in ret_location.depend:
	    ret_location.depend.append(model_transition)
        return transition
    def next(self, loc):
        ret = []
        for transition in self.goal_template.forward[loc]:       
	  next_discrete_transition = self.takeTransition(loc.zone, transition)
	  if not (next_discrete_transition is None):
	    ret.append(next_discrete_transition)
	self._forward[loc] = ret
        return ret
    def solve(self):
        waiting = []
        initial = self.goal_template.initlocation
	initial.zone = self.context.getZeroFederation()
        initial.zone.upInPlace()
        initial.zone &= initial.invariant
        initial.zone.extrapolateMaxBoundsInPlace(self.model.clock_bounds)
        if (initial.name == self.goal_location_name):
	  return False
        else:
	  initial.accepting_zone = self.context.getEmptyFederation()
	for e_ in self.next(initial):
	  logger.debug("adding to wait next initial: " + str(e_.model_transition.target) + "#" + str(e_.destination))
          waiting.append(e_)
      
        i = 0
        while waiting:
            e = waiting.pop()
            t = e.model_transition
            S = e.source
            loc = t.source
            S_ = e.destination
            loc_ = t.target
            i += 1
            if not S_ <= loc_.zone: # not included
	        logger.debug(str(i)+'Exploring new state:' + str(loc_) + "#" + str(S_) + ' [accepting=' + str(loc_.accepting)+ ']' +' (from ' + str(t) + ')')
                loc_.zone |= S_
                if not loc_.accepting:
                    for e_ in self.next(loc_):
			logger.debug("adding to wait next: " + str(e_.model_transition.target) + "#" + str(e_.destination))
			if not e_ in waiting:
			    waiting.append(e_)
                else:
		    logger.debug("adding to wait accepting: " + str(loc_) + "#" + str(S_))
		    loc_.accepting_zone = loc_.zone.copy()
		    waiting.append(e)
            else:                
                logger.debug(str(i)+'Updating :' + str(loc) + "#" + str(S) + '(from ' + str(t) + ')')
		bad  = S.up() - S # lm comment: S.zone.up() - S.zone is the zone beyond the invariant. In this zone the environnement is winning.
		for a in self.nextUncontrollable(loc):
		    bad  |= a.pred(a.model_transition.target.accepting_zone)
		#print "bad: " + str(bad)
		good  = self.context.getEmptyFederation()
		for a in self.nextControllable(loc):
		    good |= a.pred(a.destination - a.model_transition.target.accepting_zone)
		accepting_ = bad.predt(good-bad)
		accepting_ &= S
		#print accepting_

                if accepting_ != loc.accepting_zone:
		    self.strategy.append(loc)
		    for a in loc.depend:
			logger.debug("adding to wait dependance: " + str(a.target) + "#" + str(a.target.zone))
			e_ = DiscreteTransition(a.source.zone,a.target.zone,a)
			if not e_ in waiting:
			    waiting.append(e_)
                    loc.accepting_zone = accepting_
                    logger.debug('Update :' + str(loc)+ "#" + str(S) + '[accepting=' + str(loc.accepting_zone)+ ']')
		#loc_.depend.add(t)  
		if not t in loc_.depend:
		    loc_.depend.append(t)
		logger.debug('Accepting zone:' + str(loc.accepting_zone))
	    
	    if initial.accepting_zone.hasZero(): # we can't do that in the preamble of while-loop because we have to iterate at least once
		self.solved = -1
                break

	logger.info(str(i) + " states visited.")
	if self.solved >=0:
	    self.solved = 1 # game is won by the controller
        return (self.solved > 0)
    def nextControllable(self, loc):
      	logger.debug("nextControllable")
        return [t for t in self._forward[loc] if t.model_transition.controllable]
    def nextUncontrollable(self, loc):
      	logger.debug("nextUncontrollable")
        return [t for t in self._forward[loc] if not t.model_transition.controllable]

    def extractStrategy(self):
	""" Extract a strategy from the timed game by computed the set of states that belongs to a spoiling strategy that starts from the initial state.
	    Gives a more restrictive set of states than the previously computed accepting_zone.
	    However it is not used at the moment since there is a bug in the computation.
	"""
	waiting = [(self.goal_template.initlocation,self.goal_template.initlocation.accepting_zone.copy())]
	i = 0
	while waiting:
            current,domain = waiting.pop()
            i += 1
	    logger.debug('Strategy state : ' + str(current) + " " + str(domain))
	    logger.debug('###### Strategy domain: ' + str(current.strategy_domain))
	    logger.debug('###### Accepting zone: ' + str(current.accepting_zone))
	    logger.debug('###### Zone: ' + str(current.zone))
	    
	    if not domain <= current.strategy_domain:
	      current.strategy_domain |= domain
	      
	      if not current.accepting:
		bad = self.context.getEmptyFederation()
		for a in self.nextUncontrollable(current):
		  bad  |= a.pred(a.model_transition.target.accepting_zone-a.model_transition.target.strategy_domain)
		bad &= current.strategy_domain
		
		for transition in self.goal_template.forward[current]:
		  if not transition.controllable:
		    next_discrete_transition = self.takeTransition(bad,transition)
		    if not (next_discrete_transition is None):
		      next_ = next_discrete_transition.model_transition.target
		      next_discrete_transition.destination &= next_.accepting_zone
		      if not next_discrete_transition.destination.isEmpty():
			waiting.append((next_,next_discrete_transition.destination))
			logger.debug('Adding to wait : ' + str(next_)+ "#" + str(next_discrete_transition.destination))
		  else:
		    #next_discrete_transition = self.takeTransition(current.strategy_domain-bad,transition)
		    next_discrete_transition = self.takeTransition(current.strategy_domain,transition)
		    if not (next_discrete_transition is None):
     		      next_ = next_discrete_transition.model_transition.target
		      next_discrete_transition.destination &= next_.accepting_zone
		      if not next_discrete_transition.destination.isEmpty():
			waiting.append((next_,next_discrete_transition.destination))
			logger.debug('Adding to wait : ' + str(next_)+ "#" + str(next_discrete_transition.destination))
		    #elif not test is None:
     		      #print "new*******"
		      #next_ = test.model_transition.target
		      #test.destination &= next_.accepting_zone
		      #waiting.append((next_,test.destination))
		      #logger.debug('Adding to wait : ' + str(next_)+ "#" + str(test.destination))
		      #if not test.destination.isEmpty():
			#waiting.append((next_,test.destination))
			#logger.debug('Adding to wait : ' + str(next_)+ "#" + str(test.destination))
		      #print current.strategy_domain
		      #print bad
		      #print current.strategy_domain-bad
		      #print transition
		      
    def refine(self,delta_value, delta_name="delta"):
      """ Refine the spoiling strategy to extract parametrically the minimum value of delta.
      """
      self.param_context = pyppl.ParamContext(self.context.clocks,delta_name)
      self.initial_domain = self.param_context.getInitialDomain(0,delta_value)
      logger.debug("initial domain: " + str(self.initial_domain))

      if self.solved >= 0 : # the game is winning for the environnement (or not solved)
	return self.param_context.getEmptyParamFederation()
	
      for location in self.goal_template.locations:
	location.invariant = pyparamexpr.parse_param_invariant_or_guard(location.invariant_str,delta_name,self.param_context) # overwrite the dbm domain
	location.invariant &= self.initial_domain
	if(location.accepting):
	  location.accepting_param_zone = location.invariant.copy() 
	else:
	  location.accepting_param_zone = self.param_context.getEmptyParamFederation()
      for transition in self.goal_template.transitions:
	  transition.guard = pyparamexpr.parse_param_invariant_or_guard(transition.guard_str,delta_name,self.param_context)  # overwrite the dbm domain
	  transition.update = [self.param_context.getIdVariable(clock.name.split('.')[1]) for (clock,value) in transition.update] # overwrite the dbm clocks update
		
      waiting = self.strategy
              
      i = 0
      while waiting:
	current = waiting.pop(0)
	if current.accepting_zone.isEmpty():
	  continue

	logger.debug("Refining: " + str(current))
	i += 1
  
        bad = ~current.invariant
        bad &= self.initial_domain
        for transition in self.nextUncontrollable(current):
	  if(not transition.model_transition.target.accepting_param_zone.isEmpty()):
	    bad  |= transition.model_transition.target.accepting_param_zone.pred(transition.model_transition.guard & current.invariant,transition.model_transition.update)
	good  = self.param_context.getEmptyParamFederation()
	for transition in self.nextControllable(current):
	  pred_acc = transition.model_transition.target.invariant - transition.model_transition.target.accepting_param_zone
	  good |= pred_acc.pred(transition.model_transition.guard & current.invariant,transition.model_transition.update) 
	good -= bad
	good &= self.initial_domain
	
	current.accepting_param_zone = bad.predt(good)
	current.accepting_param_zone &= current.invariant
	#accepting_.pairwiseReduce()
	
	logger.debug(str(i) + " states refined;")
	logger.debug("Refined location: " + str(current) + " ; accepting domain: " + str(current.accepting_param_zone))
	if current == self.goal_template.initlocation:
	  refine_values = self.goal_template.initlocation.accepting_param_zone.hasZeroParamDomain()
	  if not refine_values.isEmpty():
	    logger.info(str(i) + " states refined;")
	    return refine_values.minimizeParam()
      logger.error("no winning value in parameter domain: " + str(self.goal_template.initlocation.accepting_param_zone))
      return None
	      
    
        
#if __name__ == '__main__':

    #model = pyuppaal.NTA.from_xml(sys.argv[1])
    
    #dbmpyuppaal.parse_nta_dbm(model)
    
    #goal_template = model.getTemplateByName(sys.argv[2].split('.')[0])
    #goal_location_name = sys.argv[2].split('.')[1]
    
    #dbmpyuppaal.parse_templates_dbm(model,[goal_template])

    #game = SafetyTimedGame(model, goal_template, goal_location_name) 
    #print game.solve()
    #game.printDot('game.dot')

    
