"""PyECDAR, a tool for timed games and timed specifications.
Copyright (C) INRIA Rennes - Louis-Marie Traonouez

Louis-Marie.Traonouez@inria.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import logging

import ppl_ta_int
import copy

logging.basicConfig(level=logging.DEBUG)
my_logger = logging.getLogger('Python-PPL-TA')

from  ppl_ta_int import ParamGuard

class ParamFederation:
  def __init__(self, context):
    if isinstance(context, ParamContext):
      self._fed = ppl_ta_int.ParamFederation(len(context.variables))
      self.context = context 
    else:
      my_logger.error("Wrong argument for ParamFederation constructor")
      assert(0)
    
  def __str__(self):
    return self.context.formatExpression(self._fed.toStr())
  __repr__ = __str__
  
  def copy(self):
    ret = self.__class__(self.context)
    ret._fed = ppl_ta_int.ParamFederation(self._fed)
    return ret
  __deepcopy__ = lambda self, memo: self.copy()
  __copy__ = copy
  
  def __and__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    assert(arg2.context == self.context)        
    ret = self.copy()
    ret._fed.add_conjunction(arg2._fed)
    return ret
  def __iand__(self, arg2):        
    assert(isinstance(arg2, ParamFederation))
    assert(arg2.context == self.context)
    self._fed.add_conjunction(arg2._fed)
    return self
    
  def __or__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    assert(arg2.context == self.context)
    ret = self.copy()
    ret._fed.add_disjunction(arg2._fed)
    return ret
  def __ior__(self, arg2):        
    assert(isinstance(arg2, ParamFederation))
    assert(arg2.context == self.context)
    self._fed.add_disjunction(arg2._fed)
    return self
  
  def __sub__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    assert(arg2.context == self.context)
    ret = self.copy()
    ret._fed.difference(arg2._fed)
    return ret
  def __isub__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    assert(arg2.context == self.context)
    self._fed.difference(arg2._fed)
    return self
    
  def __invert__(self):
    ret = self.__class__(self.context)
    ret._fed = self._fed.negation()
    return ret
    
  def addDisjunction(self,g):   
    assert(isinstance(g,ParamGuard))
    self._fed.add_disjunction(g)
    
  def addConjunction(self,g):   
    assert(isinstance(g,ParamGuard))
    self._fed.add_conjunction(g)

  def up(self):
    ret = self.copy()
    ret.upInPlace()
    return ret 
  def upInPlace(self):
    self._fed.future(1)
  
  def down(self):
    ret = self.copy()
    ret.downInPlace()
    return ret 
  def downInPlace(self):
    self._fed.past(1)

  def freeVariable(self, v_id):
    ret = self.copy()
    ret.freeVariableInPlace(v_id)
    return ret 
  def freeVariableInPlace(self, v_id):
    self._fed.free(v_id)
  
  def resetValue(self, v_id):
    ret = self.copy()
    ret.resetValueInPlace(v_id)
    return ret
  def resetValueInPlace(self, v_id):
    self._fed.reset(v_id)
  
  def post(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    assert(guard.context == self.context)
    ret = self.copy()
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    ret._fed.post(guard._fed,reset_vector)
    return ret
  def postInPlace(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    assert(guard.context == self.context)
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    self._fed.post(guard._fed,reset_vector)
   
  def pred(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    assert(guard.context == self.context)
    ret = self.copy()
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    ret._fed.pred(guard._fed,reset_vector)
    return ret
  def predInPlace(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    assert(guard.context == self.context)
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    self._fed.pred(guard._fed,reset_vector)
         
  def predt(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    assert(arg2.context == self.context)
    ret = self.__class__(self.context)
    ret._fed = ppl_ta_int.ParamFederation_predt(self._fed,arg2._fed,1)
    return ret 

  def pairwiseReduce(self):
    self._fed.pairwise_reduce()
    
  def contains (self, arg2):
    return self._fed.contains(arg2._fed)
    
  def isDisjointFrom(self, arg2):
    return self._fed.is_disjoint_from(arg2._fed)
    
  def geometricallyEquals(self, arg2):
    return self._fed.geometrically_equals(arg2._fed)
  def geometricallyCovers(self, arg2):
    return self._fed.geometrically_covers(arg2._fed)
    
  def paramProjection(self):
    ret = self.copy()
    ret._fed = self._fed.param_projection(1)
    return ret
    
  def paramProjectionInPlace(self):
    self._fed.param_projection(1)
    
  def hasZeroParamDomain(self):
    zero = self.copy()
    for i in xrange(1,len(self.context.variables)):
      zero.addConjunction(ppl_ta_int.ParamGuard(i,0,0,0,0))
    zero.paramProjectionInPlace()
    return zero
    
  def minimizeParam(self):
    found,min_n,min_d,minimum = self._fed.minimize_param(0)
    if found:
      return (min_n,min_d,minimum)
    else:
      return None

  def isEmpty(self):
    return self._fed.is_empty()
  def getSize(self):
    return self._fed.size();


class ParamContext:
  def __init__(self, clocks, param_name):
    self.variables = {}
    v_id = 0
    self.variables[param_name] = v_id
    v_id += 1
    for clock in clocks:
      self.variables[clock.name.split('.')[1]] = v_id
      v_id += 1
       
  def getEmptyParamFederation(self):
    ret = ParamFederation(self)
    ret._fed = ppl_ta_int.ParamFederation(len(self.variables),ppl_ta_int.EMPTY)
    return ret
      
  def getTautologyParamFederation(self):
    return ParamFederation(self)

  def getInitialDomain(self,min_param_value=0,max_param_value="inf"):
    ret = ParamFederation(self)
    ret.addConjunction(ppl_ta_int.ParamGuard(0,0,min_param_value,0,1)) #delta >= min_param_value
    if max_param_value < "inf":
      ret.addConjunction(ppl_ta_int.ParamGuard(0,0,max_param_value,0,-1)) #delta <= max_param_value
    for i in xrange(1,len(self.variables)):
      ret.addConjunction(ppl_ta_int.ParamGuard(i,0,0,0,1)) #x_i >= 0 (i.e. v(i)>=0+0*v(0))
    return ret
  
  def getZeroDomain(self,min_param_value=0,max_param_value="inf"):
    ret = ParamFederation(self)
    ret.addConjunction(ppl_ta_int.ParamGuard(0,0,min_param_value,0,1)) #delta >= min_param_value
    if max_param_value < "inf":
      ret.addConjunction(ppl_ta_int.ParamGuard(0,0,max_param_value,0,-1)) #delta <= max_param_value
    for i in xrange(1,len(self.variables)):
      ret.addConjunction(ppl_ta_int.ParamGuard(i,0,0,0,1)) #x_i >= 0 (i.e. v(i)>=0+0*v(0))
      ret.addConjunction(ppl_ta_int.ParamGuard(i,0,0,0,0)) #x_i <= 0 (i.e. v(i)>=0+0*v(0))
    return ret
  
  def getIdVariable(self,v_name):
    return self.variables[v_name]
    
  def formatExpression(self,expr):
    if len(self.variables)<=26:
      for v_name, v_id in self.variables.items():
	expr = expr.replace(chr(v_id+65),v_name)
      return expr
    else:
      return expr
      
