/*PyECDAR, a tool for timed games and timed specifications.
Copyright (C) INRIA Rennes - Louis-Marie Traonouez

Louis-Marie.Traonouez@inria.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.*/

#include "ppl_ta_int.h"

ParamFederation::ParamFederation(const NNC_Polyhedron& p) : Pointset_Powerset<NNC_Polyhedron>(p)
{ }

void
ParamFederation::add_disjunction(const ParamGuard& c) {
  NNC_Polyhedron ph(this->space_dimension());
  ph.add_constraint(c.guard_domain);
  this->add_disjunct(ph);
}
    

void
ParamFederation::add_disjunction(const Constraint& c)
{
  NNC_Polyhedron ph(this->space_dimension());
  ph.add_constraint(c);
  this->add_disjunct(ph);
}


void
ParamFederation::add_disjunction(const ParamFederation& fed) {
  Pointset_Powerset<NNC_Polyhedron>::const_iterator it;
  for(it = fed.begin(); it != fed.end(); it++) {
    this->add_disjunct(it->pointset());
  }
}
    

void
ParamFederation::add_conjunction(const ParamGuard& c) {
  this->add_constraint(c.guard_domain);
}
   

void
ParamFederation::add_conjunction(const Constraint& c)
{
  this->add_constraint(c);
}


void
ParamFederation::add_conjunction(const ParamFederation& fed) {
  this->intersection_assign(fed);
}
    

///////////// Negations
ParamFederation
ParamFederation::negation() const {
  ParamFederation result(this->space_dimension(),UNIVERSE);
  Pointset_Powerset<NNC_Polyhedron>::const_iterator it;
  for(it = this->begin(); it != this->end(); it++) {
    // transform polyhedron into a disjunction of the negation of its constraint
    ParamFederation inter(this->space_dimension(),EMPTY);
    Constraint_System sys = it->pointset().minimized_constraints();
    Constraint_System::const_iterator ci;
    for (ci=sys.begin(); ci!=sys.end(); ci++)
    {
      if((*ci).is_equality()) {
	Linear_Expression e;
	for (int i = (*ci).space_dimension() - 1; i >= 0; i--)
	  e += (*ci).coefficient(Variable(i)) * Variable(i);
	e += (*ci).inhomogeneous_term();
	Constraint neg1 = (e < 0);
	Constraint neg2 = (e > 0);
	inter.add_disjunction(neg1);
	inter.add_disjunction(neg2);
      }
      else {
	inter.add_disjunction(negation(*ci));
      }
    }
    // add in conjunction the intermediate result to the result
    result.add_conjunction(inter);
  }
  return result;
}


Constraint
ParamFederation::negation(const Constraint& c) {
  if (c.is_equality())
    return c; // this should be an error
  else {
    Linear_Expression e;
    for (int i = c.space_dimension() - 1; i >= 0; i--)
      e += c.coefficient(Variable(i)) * Variable(i);
    e += c.inhomogeneous_term();
    Constraint neg = c.is_strict_inequality() ? (e <= 0) : (e < 0);
    return neg;
  }
}


void
ParamFederation::difference(const ParamFederation& fed) {
  this->add_conjunction(fed.negation());
  this->pairwise_reduce();
}
    

////////////// Timed Automata specific 

void
ParamFederation::future(int num_param) {
  int i;
  ParamFederation q (this->space_dimension());
  for (i = 0; i < num_param; i++)
  {
    q.add_conjunction(Constraint(Variable(i)==0));
  }
  for (; i < this->space_dimension(); i++)
  {
    q.add_conjunction(Constraint(Variable(i)==1));
  }
  this->time_elapse_assign(q);
}
    

void
ParamFederation::past(int num_param) {
  int i;
  ParamFederation q (this->space_dimension());
  for (i = 0; i < num_param; i++)
  {
    q.add_conjunction(Constraint(Variable(i)==0));
  }
  for (; i < this->space_dimension(); i++)
  {
    q.add_conjunction(Constraint(Variable(i)==-1));
  }
  this->time_elapse_assign(q);
  for (; i-- > num_param;)
  {
    this->add_conjunction(Constraint(Variable(i)>=0));
  }
}
    

void
ParamFederation::free(int var) {
  this->unconstrain(Variable(var));
  this->add_constraint(Variable(var)>=0);
}
    
void
ParamFederation::reset(int var) {
  this->unconstrain(Variable(var));
  this->add_constraint(Variable(var)==0);
}
    
void
ParamFederation::post(const ParamFederation& guard, const std::vector<int>& reset) {
  this->add_conjunction(guard);
  for(int j=0; j < reset.size(); j++) {
    this->reset(reset[j]);
  }
}

void
ParamFederation::pred(const ParamFederation& guard, const std::vector<int>& reset) {
  for(int i = 0; i < reset.size(); i++) {
    this->add_constraint(Variable(reset[i])==0);
    this->unconstrain(Variable(reset[i]));
  }
  this->add_conjunction(guard);
}


ParamFederation
ParamFederation::predt(const ParamFederation& good,const ParamFederation& bad,int num_param)
{
  ParamFederation result(good.space_dimension(),EMPTY);
  
  if(good.space_dimension() == bad.space_dimension()) {
    if(bad.empty()) { // no disjunctions
      result = ParamFederation(good);
      result.past(num_param);
    }
    else {
      Pointset_Powerset<NNC_Polyhedron>::const_iterator goodi,badj;
      for(goodi = good.begin(); goodi != good.end(); goodi++) {
	ParamFederation inter(good.space_dimension(),UNIVERSE);
	for(badj = bad.begin(); badj != bad.end(); badj++) {
	  inter.add_conjunction(predt(goodi->pointset(),badj->pointset(),num_param));
	}
	result.add_disjunction(inter);
      }
    }
  }
  return result;  
}


ParamFederation
ParamFederation::predt(const NNC_Polyhedron& good,const NNC_Polyhedron& bad,int num_param)
{
  ParamFederation bad_past = ParamFederation(bad);
  bad_past.past(num_param);
  ParamFederation good_past = ParamFederation(good);
  good_past.past(num_param);
  
  ParamFederation result = ParamFederation(good);
  result.add_conjunction(bad_past);
  
  if(not result.is_empty()) {
    return good_past;
  }
  else {
    good_past.difference(bad_past);
    return good_past;
  }
  
//   ParamFederation good_past = ParamFederation(good);
//   good_past.past(num_param);
//   ParamFederation bad_past = ParamFederation(bad);
//   bad_past.past(num_param);
//     
//   good_past.difference(bad_past);
//   bad_past.add_conjunction(good);
//   //   bad_past.difference(bad); not necessary for safety since good inter bad is empty
//   bad_past.past(num_param);
//   good_past.add_disjunction(bad_past);
//     
//   return good_past;
}


////////////// Parameter specific 

void
ParamFederation::param_projection(int num_param) {
  this->remove_higher_space_dimensions(num_param);
}


void
ParamFederation::minimize_param(int id_param,bool* found,int* min_n,int* min_d,bool* minimum) const {
  Variable delta(id_param);
  Coefficient min_n_coeff;
  Coefficient min_d_coeff;
  Linear_Expression delta_expr(delta);
  
  *found = this->minimize(delta_expr,min_n_coeff,min_d_coeff,*minimum);
  
  stringstream out_n, out_d;
  out_n << min_n_coeff;
  out_d << min_d_coeff;
  *min_n = atoi(out_n.str().c_str());
  *min_d = atoi(out_d.str().c_str());
}


////////////// Printing
string
ParamFederation::toStr() const {
  stringstream out;
  out << (*this);
  return out.str();
}

