/*PyECDAR, a tool for timed games and timed specifications.
Copyright (C) INRIA Rennes - Louis-Marie Traonouez

Louis-Marie.Traonouez@inria.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.*/

#ifndef PPL_TA_INT_H
#define PPL_TA_INT_H

#include <iostream>
#include <string>
#include <vector>

#include <ppl.hh>

using namespace std;
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;

    
/*! \brief A parametric guard with one parameter Delta.*/
class ParamGuard {
  friend class ParamFederation;

  private:
    Constraint guard_domain;

    ParamGuard () : guard_domain(Variable(0)==0) { }
        
  public:
    
    /*! \brief x_i - x_j --sign-- c + d*Delta.
     * sign= -2:<,-1:<=,0:==,1:>=,2:>
     */
    ParamGuard (int i, int j, int delta, int c, int d, int sign) : guard_domain(Variable(0)==0) {
      switch(sign) {
	case -2:
	  guard_domain = Constraint(Variable(i) - Variable(j) < c + d*Variable(delta));
	  break;
	case -1:
	  guard_domain = Constraint(Variable(i) - Variable(j) <= c + d*Variable(delta));
	  break;
	case 0:
	  guard_domain = Constraint(Variable(i) - Variable(j) == c + d*Variable(delta));
	  break;
	case 1:
	  guard_domain = Constraint(Variable(i) - Variable(j) >= c + d*Variable(delta));
	  break;
	case 2:
	  guard_domain = Constraint(Variable(i) - Variable(j) > c + d*Variable(delta));
	  break;
      }
    }

    /*! \brief x_i < c + d*Delta.
     * sign= -2:<,-1:<=,0:==,1:>=,2:>
     */
    ParamGuard (int i, int delta, int c, int d, int sign) : guard_domain(Variable(0)==0) {
      switch(sign) {
	case -2:
	  guard_domain = Constraint(Variable(i) < c + d*Variable(delta));
	  break;
	case -1:
	  guard_domain = Constraint(Variable(i) <= c + d*Variable(delta));
	  break;
	case 0:
	  guard_domain = Constraint(Variable(i) == c + d*Variable(delta));
	  break;
	case 1:
	  guard_domain = Constraint(Variable(i) >= c + d*Variable(delta));
	  break;
	case 2:
	  guard_domain = Constraint(Variable(i) > c + d*Variable(delta));
	  break;
      }
    }
    
    /*! \brief Print the guard domain intro a string. */
    string toStr() const {
      stringstream out;
      out << guard_domain;
      return out.str();
    }
};


/*! \brief An implementation of parametric federations for TA based on polyhedra using the PPL library.
 *  We assume the parameters are in the first variables of the domain. All others variables are clocks.
 */	
class ParamFederation : public Pointset_Powerset<NNC_Polyhedron>
{
  public:
    /*! \brief Empty constructor for an EMPTY domain. */	
    ParamFederation(int num_dimensions=0,Degenerate_Element kind=UNIVERSE) : Pointset_Powerset<NNC_Polyhedron>(num_dimensions,kind) { }
    
    /*! \brief Constructor with an initial polyhedron. */	
    ParamFederation(const NNC_Polyhedron& p);
    
    /*! \brief Copy constructor. */	
    ParamFederation(const ParamFederation& fed) : Pointset_Powerset<NNC_Polyhedron>(fed) { }

    /*! \brief Number of dimensions. */	
    int size() const {
      return this->space_dimension();
    }
    
    //////////// Add new disjunctions:
    /*! \brief Add a parametric constraint in disjunction. */	
    void add_disjunction(const ParamGuard& c);
    
    /*! \brief Add a constraint in disjunction. */	
    void add_disjunction(const Constraint& c);

    /*! \brief Add a formula in disjunction. 
      *  Add each polyhedron in fed in disjunction.
      */	
    void add_disjunction(const ParamFederation& fed);
    
    //////////// Add in conjunction:
    /*! \brief Add a parametric constraint in conjunction. */	
    void add_conjunction(const ParamGuard& c);
    
    /*! \brief Add a constraint in conjunction. */	
    void add_conjunction(const Constraint& c);

    /*! \brief Add a formula in conjunction. */
    void add_conjunction(const ParamFederation& fed);
    
    ///////////// Negations
    /*! \brief Compute the negation of this formula. */
    ParamFederation negation() const;
    
    /*! \brief Compute the negation of a single constraint. */	
    static Constraint negation(const Constraint& c);

   
    /*! \brief Assign to this the difference between this and the federation passed in parameter.
      *  P\Q = P and not Q
      */
    void difference(const ParamFederation& fed);
    
    
    ////////////// Timed Automata specific 
    /*! \brief Assign to this the future of this.
     * The parameters in the first num_param variables are constant in this transformation.
     */
    void future(int num_param);
    
    /*! \brief Assign to this the past of this. 
     * The parameters in the first num_param variables are constant in this transformation.
     */
    void past(int num_param);
    
    /*! \brief Free a variable to all positive values. */
    void free(int var);
    
    /*! \brief Reset to zero the variable given in parameter. */
    void reset(int var);    

    /*! \brief Compute and assign to this the post states of this domain through an edge with guard and reset cloks. */
    void post(const ParamFederation& guard, const std::vector<int>& reset);

    /*! \brief Compute and assign to this the pre states of this domain through an edge with guard and reset cloks. */
    void pred(const ParamFederation& guard, const std::vector<int>& reset);
    
    /*! \brief Compute the time predecessors of good that avoid bad. 
     *  Good and bad are federations. Compute a union of intersection of predt applied between two polyhedra in the federation.
     */
    static ParamFederation predt(const ParamFederation& good,const ParamFederation& bad,int num_param);
    
    ////////////// Parameter specific 
        
    /*! \brief Projection on the set of parameters. */
    void param_projection(int num_param);
    
    /*! \brief Return the infinimum value of the parameter over the domain.
     *  - id_param is the variable index of the parameter 
     *  - min_n is the result numerator
     *  - min_d is the result denumerator
     *  - bool is true if the result is also a minimum value
     *  - found is true if a infinimum was found
     */
    void minimize_param(int id_param, bool* found, int* min_n, int* min_d, bool* minimum) const;
    
    ////////////// Printing
    /*! \brief Print the domain intro a string. */
    string toStr() const;


  private:
    /*! \brief Compute the time predecessors of good that avoid bad.
     *  Good and bad are convex polyhedron.
     */
    static ParamFederation predt(const NNC_Polyhedron& good,const NNC_Polyhedron& bad,int num_param);
    
};

#endif

